#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifndef nothread
  #include <pthread.h>
#endif
#include <sys/time.h>
#include <unistd.h>
#include <sys/poll.h>

#include "libcomm.h"



#ifndef nothread
/**
   This is a part of block_call() and must not called from the
   user. This function is the thread which is started from
   block_call() and runs in background.

   @param libcommpthreads (struct LIBCOMMPTHREADS *) holds pointers to
   the functions to be call, fd (socket discriptor) and id.
*/
void thread1(struct LIBCOMMPTHREADS *libcommpthreads) {
  char *buf;
  unsigned int type;
  unsigned int size;

  while (1) {
    // try to receive a datablock...
    buf = block_receive(libcommpthreads -> fd, &type, NULL, &size, 0, 
			libcommpthreads -> term);
    // failed: call block_call_term() and terminate thread
    if (buf == NULL) {
      libcommpthreads -> block_call_term(libcommpthreads -> fd,
					 libcommpthreads -> id);
      break;
    }
    // datablock OK: call block_call_do(), after this wait for the
    // next datablock
    libcommpthreads -> block_call_do(libcommpthreads -> fd,
				     libcommpthreads -> id,
				     type, buf, size,
				     libcommpthreads -> term);
  }
  pthread_exit(NULL);
}


/**
   Start in thread, wait for a datablock and call the function
   block_call_do() if one was received or block_call_term() when the
   connection terminates.

   @param fd (int) descriptor of socket

   @param id (int) arbitrary id of background process / thread

   @param term (int) 0: do not terminate the buffer, 1: terminate the
   buffer by appending a 0x00.

   @param block_call_do(int fd, int id, unsigned int type, char *buf,
   unsigned int size, int term) (function) this function is called if
   a datablock was received. fd, id and term are the same as in
   block_call(). type describes the type of the received datablock,
   buf is a pointer to this datablock and size is the number of bytes
   of the datablock

   @param block_call_term(int fd, int id) (function) this function is
   called if the connection terminates. fd and id are the same as in
   block_call().

   @return If all right zero otherwise non zero.
*/
int block_call(int fd, int id, int term,
	   void (*block_call_do)(int fd, int id, unsigned int type,
				 char *buf, unsigned int size,
				 int term),
	   void (*block_call_term)(int fd, int id)) {

  // allocate memory for thread configuration
  struct LIBCOMMPTHREADS *libcommpthreads; 
  libcommpthreads = (struct LIBCOMMPTHREADS *)
    malloc(sizeof(struct LIBCOMMPTHREADS));
  if (libcommpthreads == NULL) {
    perror("malloc()");
    return -1;
  }

  // store all necessary data in it
  libcommpthreads -> fd = fd;
  libcommpthreads -> id = id;
  libcommpthreads -> block_call_do = block_call_do;
  libcommpthreads -> block_call_term = block_call_term;

  // starting thread
  pthread_attr_init(&(libcommpthreads -> thrd_1_attr));
  return pthread_create(&(libcommpthreads -> thrd_1),
			&(libcommpthreads -> thrd_1_attr),
			(void *) thread1, libcommpthreads);
}

#endif



/**
   This function tests if new data is available to read on a stream.

   @param fd (int) discriptor of stream to test

   @return (int) 1: Data to read; 0: No data to read
*/
int block_ifdata(int fd) {
  struct pollfd polld;

  polld.fd = fd;
  polld.events = POLLIN | POLLPRI;

  if (poll(&polld, 1, 0)) {
    return 1;
  }
  return 0;
}


/**
   Test if is there data available on the socket and read it in a
   buffer if there is. WARNING: The integers (type and size; excluding
   fd) are only 16 bit values (0 - 65535).

   @param fd (int) descriptor of socket

   @param type (unsigned int *) pointer to integer, this value can be
   used as buyer's option

   @param buf (char *) buffer for datablock. Memory will be allocated
   if this parameter is equal to null.

   @param size (unsigned int *) pointer to integer in which the size
   of the received datablock is saved.

   @param maxsize (unsigned int *) describes size of buf. This
   parameter will be ignored if bus is equal to null.

   @param term (int) 0: do not terminate the buffer, 1: terminate the
   buffer by appending a 0x00.

   @return (char *) pointer to buffer which contains the received
   datablock; NULL if fail; 1 if no data available.

*/
char *block_receive_poll(int fd, unsigned int *type, char *buf,
			 unsigned int *size, unsigned int maxsize,
			 int term) {
  // new data available
  if (block_ifdata(fd)) {
    return block_receive(fd, type, buf, size, maxsize, term);
  } else {
  // no new data available
    return (char *) 1L;
  }
}



/**
   Receive a block (composition of: type, size of datablock and
   datablock) from a socket. WARNING: The integers (type and size;
   excluding fd) are only 16 bit values (0 - 65535).

   @param fd (int) descriptor of socket   

   @param type (unsigned int *) pointer to integer, this value can be
   used as buyer's option

   @param buf (char *) buffer for datablock. Memory will be allocated
   if this parameter is equal to null.

   @param size (unsigned int *) pointer to integer in which the size
   of the received datablock is saved.

   @param maxsize (unsigned int *) describes size of buf. This
   parameter will be ignored if bus is equal to null.

   @param term (int) 0: do not terminate the buffer, 1: terminate the
   buffer by appending a 0x00.

   @return (char *) pointer to buffer which contains the received
   datablock; NULL if fail.

*/
char *block_receive(int fd, unsigned int *type, char *buf,
		    unsigned int *size, unsigned int maxsize, 
		    int term) {
  // do not trust any user!
  if (term > 1) {
    term = 1;
  }
  if (term < 0) {
    term = 0;
  }

  // receiving type
  if (block_receive_integer(fd, type) < 0) {
    return NULL;
  }
  // receiving size
  if (block_receive_integer(fd, size) < 0) {
    return NULL;
  }

  if (buf == NULL) {
    if ((buf = (char *) malloc(*size + term)) == NULL) {
      perror("malloc()");
      return NULL;
    }
  } else {
    if ((*size + term) > maxsize) {
      fprintf(stderr, "Try to receive more than fit in the buffer\n");
      return NULL;
    }
  }
  // receiving data
  if (block_receive_nbytes(fd, buf, *size) < 0) {
    return NULL;
  }

  if (term) {
    buf[*size] = 0;
  }

  return buf;
}




/**
   Receive an integer (two bytes; 16Bit) from the socket. 

   @param fd (int) descriptor of socket

   @param recvi (unsigned int *) pointer to integer in which the
   received integer is saved.

   @return (int) 2: OK; -1: fail
*/
int block_receive_integer(int fd, unsigned int *recvi) {
  int i, r;
  //  unsigned int recvi;
  int sizeofint = 2; /* sizeof(int); */

  // reset value
  *recvi = 0;

  // receive value
  if (recv(fd, ((char *) recvi), sizeofint, MSG_WAITALL) != sizeofint) {
    perror("recv()");
    return -1;
  }

  return 2; //recvi;
}




/**
   Receive n bytes from socket.

   @param fd (integer) descriptor of socket

   @param buf (char *) buffer for saving the received bytes

   @param n (integer) number of bytes to receive

   @return (integer) n: OK; -1 fial
*/
int block_receive_nbytes(int fd, char *buf, int n) {
  int i, r;
  unsigned int recvi;
  int sizeofint = 2; /* sizeof(int); */

  // receive n bytes to buffer
  if (recv(fd, buf, n, MSG_WAITALL) != n) {
    perror("recv()");
    return -1;
  }

  return n;
}
