#include <stdio.h>
#include "md5.h"
#include "libcomm.h"



/**
   Free the memory space which is used by an AUTHINFO structure.

   @param (struct AUTHINFO *) pointer to structure to destroy.
*/
void free_authinfo(struct AUTHINFO *destroy) {
  free(destroy -> netname);
  free(destroy -> name);
  free(destroy -> passwd);
  free(destroy -> keyencrypt);
  free(destroy -> keydecrypt);
  free(destroy);
}



/**
   Do both side authentification. This function is usually called
   just after a socket stream is established. The function must be
   called on both sides. Both sides following these steps:

   1. get auth info ([login] name, passwd) by using getauthinfo() from
   name or netname for remote login

   2. generate random numbers

   3. exchange (first send, then receive) login names

   4. exchange random numbers

   5. calculate md5 checksum over the random numbers (received from other
   side) and the remote passwd.

   6. exchange md5 checksums

   7. get auth info from name (received from other side) for local login

   8. calculate md5 checksum over the local random numbers and the
   local passwd.

   9. check login -- compare the received md5sum (6.) with the
   generated one (8.); send acknowledgement
   
   10. receive remote acknowledgement

   11. return suitable values

   @param fd (int) describes the socket on which the authentication
   has to be done

   @param netname (char *) use netname to resolve [login] name and
   passwd of the remote machine (NULL: not specified)

   @param netname (char *) use [login] name to resolve passwd of the
   remote machine (NULL: not specified; both NULL use first entry in
   file, see getauthinfo())

   @param plocallogin (struct AUTHINFO **) (pointer to pointer to an
   AUTHINFO struct) in this (double pointed) struct the local authinfo
   will be loaded, if the parameter is not null.

   @param premotelogin (struct AUTHINFO **) in this (double pointed)
   struct the remote authinfo will be loaded, if the parameter is not
   null.

   @return (int) 0: Authentication/Login OK; -1: remote login error;
   -2: login error on both sides; -3: local login error; -4: other
   (network) error; -5: cannot load remote auth info; -6: cannot load
   local auth info; 
*/
int socket_md5auth(int fd, char *netname, char *name,
		   struct AUTHINFO **plocallogin,
		   struct AUTHINFO **premotelogin) {
  char rstr0[authrandomstringsize], rstr1[authrandomstringsize];
  char rstr0sum[35], rstr1sum[35];

  int otype;
  char *oname;
  int onamesize;
  char *randblock;
  char *orandblock;
  int orandblocksize;

  struct MD5Context context;
  unsigned char md5_prs[16];
  unsigned char omd5_prs[16];

  int login_ok = 0;

  struct AUTHINFO *locallogin;
  // 1.
  struct AUTHINFO *remotelogin = getauthinfo(netname, name);
  if (remotelogin == NULL) {
    return -5;
  }
  if (premotelogin != NULL) {
    *premotelogin = remotelogin;
  }
  
  // 2. generate random block for local login
  if ((randblock = block_random(NULL, authrandomstringsize))
      == NULL) {
    if (premotelogin == NULL) {
      free_authinfo(remotelogin);
    }
    return -4;
  }

  // 3. exchange login name
  if (block_send(fd, authmessagetype, remotelogin -> name,
		 strlen(remotelogin -> name)) <= 0) {
    if (premotelogin == NULL) {
      free_authinfo(remotelogin);
    }
    free(randblock);
    return -4;
  }
  oname = block_receive(fd, &otype, NULL, &onamesize, 0, true);
  if ((oname == NULL) ||
      (otype != authmessagetype)) {
    if (premotelogin == NULL) {
      free_authinfo(remotelogin);
    }
    free(randblock);
    return -4;
  }


  // 4. exchange random block
  if (block_send(fd, authmessagetype, randblock,
		 authrandomstringsize)
      <= 0) {
    if (premotelogin == NULL) {
      free_authinfo(remotelogin);
    }
    free(randblock);
    free(oname);
    return -4;
  }
  orandblock = block_receive(fd, &otype, NULL, 
			     &orandblocksize, 0, false);
  if ((orandblock == NULL) ||
      (otype != authmessagetype)) {
    if (premotelogin == NULL) {
      free_authinfo(remotelogin);
    }
    free(randblock);
    free(oname);
    free(orandblock);
    return -4;
  }

  // 5. calculate md5 checksum over the random numbers (received from
  // other side) and the remote passwd.
  MD5Init(&context);
  MD5Update(&context, orandblock, orandblocksize);
  MD5Update(&context, remotelogin -> passwd,
	    strlen(remotelogin -> passwd));
  MD5Final(md5_prs, &context);

  // 6. exchange md5 checkumms
  if (block_send(fd, authmessagetype, md5_prs, 16) <= 0) {
    if (premotelogin == NULL) {
      free_authinfo(remotelogin);
    }
    free(randblock);
    free(oname);
    free(orandblock);
    return -4;
  }
  if ((block_receive(fd, &otype, omd5_prs, &orandblocksize,
		     16, false)
      == NULL) ||
    (otype != authmessagetype) ||
    (orandblocksize != 16)) {
    if (premotelogin == NULL) {
      free_authinfo(remotelogin);
    }
    free(randblock);
    free(oname);
    free(orandblock);
    return -4;
  }
  usleep(1);

  // 7. get auth info from name (received from other side) for local login
  locallogin = getauthinfo(NULL, oname);
  if (locallogin == NULL) {
    if (premotelogin == NULL) {
      free_authinfo(remotelogin);
    }
    free(randblock);
    free(oname);
    free(orandblock);
    return -6;
  }
  if (plocallogin != NULL) {
    *plocallogin = locallogin;
  }

  // 8. calculate md5 checksum over the local random numbers and the
  // local passwd.
  MD5Init(&context);
  MD5Update(&context, randblock, authrandomstringsize);
  MD5Update(&context, locallogin -> passwd,
	    strlen(locallogin -> passwd));
  MD5Final(md5_prs, &context);
  
  // 9. check login -- compare the received md5sum (6.) with the
  // generated one (8.); send acknowledgement
  if (strncmp(md5_prs, omd5_prs) == 0) {
    login_ok = 1;
    if (block_send(fd, authmessagetype, "OK", 2) <= 0) {
      if (plocallogin == NULL) {
	free_authinfo(locallogin);
      }
      if (premotelogin == NULL) {
	free_authinfo(remotelogin);
      }
      free(randblock);
      free(oname);
      free(orandblock);
      return -4;
    }
  } else {
    if (block_send(fd, authmessagetype, "FAIL", 4) <= 0) {
      if (plocallogin == NULL) {
	free_authinfo(locallogin);
      }
      if (premotelogin == NULL) {
	free_authinfo(remotelogin);
      }
      free(randblock);
      free(oname);
      free(orandblock);
      return -4;
    }
  }

  // 10. receive remote acknowledgement
  free(orandblock);
  orandblock = block_receive(fd, &otype, NULL, &orandblocksize,
			     0, true);
  if ((orandblock == NULL) ||
      (otype != authmessagetype) ||
      (orandblocksize != 2) ||
      (strcmp(orandblock, "OK") != 0)) {
    if (login_ok) {
      if (plocallogin == NULL) {
	free_authinfo(locallogin);
      }
      if (premotelogin == NULL) {
	free_authinfo(remotelogin);
      }
      free(randblock);
      free(oname);
      return -1;
    } else {
      if (plocallogin == NULL) {
	free_authinfo(locallogin);
      }
      if (premotelogin == NULL) {
	free_authinfo(remotelogin);
      }
      free(randblock);
      free(oname);
      return -2;
    }
  }


  if (plocallogin == NULL) {
    free_authinfo(locallogin);
  }
  if (premotelogin == NULL) {
    free_authinfo(remotelogin);
  }
  free(randblock);
  free(oname);
  free(orandblock);
  
  if (!login_ok) {
    return -3;
  }

  // all right!
  return 0;
}



/**
   Load authentication informations (netname, name, passwd,
   keyencrypt, keydecrypt) from authfile.

   @param netname (char *) specify the network name (may IP). NULL not
   specified.

   @param name (char *) specity the login name. NULL not
   specified.

   @return (struct AUTHINFO *) the first entry from authfile which
   matches network name OR login name. If both values are NULL, the
   first entry of the authfile is given back.
*/
struct AUTHINFO *getauthinfo(char *netname, char *name) {
  // Descriptor for authfile
  FILE *f;
  // buffer for reading one line of the authfile
  char buf[authfilemaxlinelenght];
  // number of fields in the authfile
  int fields = 5;
  // pointer buffer for the five parts of the line
  char *bufsplit[fields];
  // char **bufsplit;
  // control variable, count variable for field
  int i, j;
  struct AUTHINFO *load;
  // temporary pointer
  char *s;

  //  bufsplit = (char **) malloc(sizeof(char *) * fields);

  // allocate memory for auth-structure
  load = (struct AUTHINFO *) malloc(sizeof(struct AUTHINFO));
  if (load == NULL) {
    perror("malloc(sizeof(struct AUTHINFO))");
    return NULL;
  }

  // open authfile
  if ((f = fopen(authfile0, "ro")) == NULL) {
    perror(authfile0);
    if ((f = fopen(authfile1, "ro")) == NULL) {
      perror(authfile0);
      free(load);
      return NULL;
    }
  }

  // read as long as ther is no more data
  while (!feof(f)) {
    // read one line
    fgets(buf, authfilemaxlinelenght - 1, f);

    // split the line into it five components
    j = 0;
    bufsplit[j++] = buf;
    //    load -> netname = buf;
    for (i = 0; i < authfilemaxlinelenght; i++) {
      if (buf[i] == authfilefieldseperator) {
	buf[i] = 0;
	if (j == fields) {
	  break;
	}
	bufsplit[j++] = buf + i + 1;
      }
    }

    // if this the right entry? Compare with parameter.
    if (
	((   name != NULL) && (strcmp(   name, bufsplit[1]) == 0))
	||
	((netname != NULL) && (strcmp(netname, bufsplit[0]) == 0))
	||
	((netname == NULL) && (name == NULL))
       ) {
      
      // allocate Memory
      for (i = 0; i < fields; i++) {
	s = NULL;
	s = (char *) malloc(strlen(bufsplit[i]) + 1);
	if (s == NULL) {
	  perror("malloc()");
	  free(load);
	  return NULL;
	}
	strcpy(s, bufsplit[i]);
	bufsplit[i] = s;
      }

      // store the pointers in the struct
      load -> netname =    bufsplit[0];
      load -> name =       bufsplit[1];
      load -> passwd =     bufsplit[2];
      load -> keyencrypt = bufsplit[3];
      load -> keydecrypt = bufsplit[4];

      // return the pointer to this struct
      return load;
    }
  }
  free(load);
  return NULL;
}
