#include <netinet/in.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>



/**
  Connect a TCP-stream to a server (Client side). Creates a socket and
  connect it over a TCP-stream to the specified port on the specified
  server.

  @param host a string (char *) which specifies the name or the
  IP-address of the server.

  @param port an integer which specifies the port on the server.

  @return The File Descriptor (FD) which allows access to the
  TCP-stream-socket or -1 if the connection fails.
*/
int socket_connect(char *host, int port) {
  // FD of the new socket to the TCP-stream
  int sock;
  // The IP address in binary form
  in_addr_t inaddr;
  // address information to connect other side (syscall: connect() )
  struct sockaddr_in ad;
  // contains the result of the resolution of a network-name.
  struct hostent *hp;
  
  // make ensure that the memory is initiated
  memset(&ad, 0, sizeof(ad));
  // address family: AF_INET: IPv4 Internet protocols
  ad.sin_family = AF_INET;
  // Try to convert the given IP-address into binary data...
  inaddr = inet_addr(host);
  if (inaddr != INADDR_NONE) {
    // if the IP address was converted copy it in the parameter
    // structure (ad) for later use
    memcpy(&ad.sin_addr, &inaddr, sizeof(inaddr));
  } else {
    // if this is not possible (the name and not the IP address is
    // given), try to resolve the name to a binary IP address
    hp = gethostbyname(host);
    if (hp == NULL) {
      // name cannot resolved, return error
      perror("gethostbyname()");
      return -1;
    }
    // copy address in the parameter structure (ad) for later use
    memcpy(&ad.sin_addr, hp->h_addr, hp->h_length);
  }
  // convert and copy port-number in the parameter structure (ad) for
  // later use
  ad.sin_port = htons(port);
  // create a socket
  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    // cannot created socket, return error
    perror("socket()");
    return -1;
  }
  // connect the socket over an TCP-stream to the port and the server,
  // which are stored in ad.
  if (connect(sock, (struct sockaddr *) &ad, sizeof(ad)) < 0) {
    // connection is not possible, return error
    perror("connect()");
    return -1;
  }
  // all right, socket is connected an can be used. Return the FD as
  // reference for use.
  return sock;
}
